use Test::More tests => 60;
use strict;

use DE_EPAGES::Core::API::Script qw ( RunScript );
use DE_EPAGES::Database::API::Connection qw ( GetTestSite RunOnStore );
use DE_EPAGES::Object::API::Factory qw (
    ExistsClassByAlias
    LoadClassByAlias
    LoadRootObject
);
use DE_EPAGES::Permission::API::AttributePermission qw (
    GetReadableAttributes
    GetWritableAttributes
    IsAttributeReadable
    IsAttributeWritable
);
use DE_EPAGES::Permission::API::Constants qw (
    $ATTRIBUTE_PERMISSION_DENY
    $ATTRIBUTE_PERMISSION_READ
    $ATTRIBUTE_PERMISSION_WRITE
);
use DE_EPAGES::Permission::API::Permission qw ( Everyone );
use DE_EPAGES::Permission::API::Table::AttributePermission qw (
    DeleteAttributePermission
    ExistsAttributePermission
    InsertAttributePermission
    UpdateAttributePermission
);
use DE_EPAGES::Core::API::Error qw ( ExistsError GetError );

use Readonly;
Readonly::Scalar my $CLASSNAME   => 'FancyClass';
Readonly::Scalar my $TESTUSER    => 'TestUser';
Readonly::Scalar my $SYSTEMGROUP => 'SystemGroup';

sub TestAttributePermissions {
    my ($Slogan, $RESTGroup) = @_;

    my $hSloganPermissionAttributes =  {
        'AttributeID' => $Slogan->id,
        'TrusteeID'   => $RESTGroup->id,
        'Permission'  => $ATTRIBUTE_PERMISSION_READ,
    };

    ok( !ExistsAttributePermission($hSloganPermissionAttributes),
        'permissions for the attribute ' . $Slogan->alias . ' are undefined' );
    ok( !IsAttributeReadable($Slogan, $RESTGroup),
        'attribute ' . $Slogan->alias . ' is not readable, yet' );

    InsertAttributePermission($hSloganPermissionAttributes);
    ok( ExistsAttributePermission($hSloganPermissionAttributes),
        'attribute permissions set for trustee ' . $RESTGroup->alias );
    ok( IsAttributeReadable($Slogan, $RESTGroup),
        'attribute ' . $Slogan->alias . ' is readable' );
    ok( !IsAttributeWritable($Slogan, $RESTGroup),
        'attribute ' . $Slogan->alias . ' is not writable' );

    # update attribute permission by using 'write' instead of 'read'
    $hSloganPermissionAttributes->{'Permission'} = $ATTRIBUTE_PERMISSION_WRITE;
    UpdateAttributePermission($hSloganPermissionAttributes);
    ok( IsAttributeWritable($Slogan, $RESTGroup),
        'attribute ' . $Slogan->alias . ' is writable' );

    # remove attribute permission
    DeleteAttributePermission($hSloganPermissionAttributes);
    ok( !ExistsAttributePermission($hSloganPermissionAttributes),
        'slogan attribute permissions were deleted' );

    return;
}

sub TestReadOnlyAttributePermissions {
    my ($ReadonlyAttribute, $RESTGroup) = @_;

    my $hReadonlyPermissionAttributes =  {
        'AttributeID' => $ReadonlyAttribute->id,
        'TrusteeID'   => $RESTGroup->id,
        'Permission'  => $ATTRIBUTE_PERMISSION_WRITE,
    };

    # try to add 'write' permissions to a read-only attribute
    eval {
        InsertAttributePermission($hReadonlyPermissionAttributes);
    };
    ok( ExistsError() && GetError()->code eq 'Permission_Not_Allowed',
        'read-only attribute is not allowed to be writable (insert)' );

    # insert correct permissions
    $hReadonlyPermissionAttributes->{'Permission'} = $ATTRIBUTE_PERMISSION_READ;
    InsertAttributePermission($hReadonlyPermissionAttributes);
    ok( ExistsAttributePermission($hReadonlyPermissionAttributes),
        'read permissions for read-only attribute were inserted' );

    # try to update a read-only attribute by changing the permission to 'write'
    eval {
        $hReadonlyPermissionAttributes->{'Permission'} = $ATTRIBUTE_PERMISSION_WRITE;
        UpdateAttributePermission($hReadonlyPermissionAttributes);
    };
    ok( ExistsError() && GetError()->code eq 'Permission_Not_Allowed',
        'read-only attribute is not allowed to be writable (update)' );

    # remove attribute permission
    DeleteAttributePermission($hReadonlyPermissionAttributes);
    ok( !ExistsAttributePermission($hReadonlyPermissionAttributes),
        'readonly attribute permissions were deleted' );

    return;
}

sub TestAttributePermissionInheritance {
    my ($Slogan, $hGroups) = @_;

    $hGroups->{'Group'}->addMember($hGroups->{'User'});
    $hGroups->{'SystemGroup'}->addMember($hGroups->{'Group'});

    # set permissions only on the top most group
    my $hSystemGroupAttributePermissions =  {
        'AttributeID' => $Slogan->id,
        'TrusteeID'   => $hGroups->{'SystemGroup'}->id,
        'Permission'  => $ATTRIBUTE_PERMISSION_READ,
    };
    InsertAttributePermission($hSystemGroupAttributePermissions);

    # verify that for each trustee, the attribute is readable
    foreach my $Group ( keys %{ $hGroups } ) {
        ok( IsAttributeReadable($Slogan, $hGroups->{$Group}),
            'attribute ' . $Slogan->alias . ' is readable for group ' . $Group );
    }

    # set the attribute permission to writable for only the group
    my $GroupAttributePermissions =  {
        'AttributeID' => $Slogan->id,
        'TrusteeID'   => $hGroups->{'Group'}->id,
        'Permission'  => $ATTRIBUTE_PERMISSION_WRITE,
    };
    InsertAttributePermission($GroupAttributePermissions);

    # verify that the user has now write-access to the attribute in question
    ok( IsAttributeWritable($Slogan, $hGroups->{'User'}),
        'attribute ' . $Slogan->alias . ' is now writable by the user' );

    # the base group should still have write read permissions
    # due to the depth of inheritance
    ok( !IsAttributeWritable($Slogan, $hGroups->{'SystemGroup'}),
        'attribute ' . $Slogan->alias . ' is not writable by the system group' );

    # remove groups and attribute permissions
    $hGroups->{'Group'}->removeMember($hGroups->{'User'});
    $hGroups->{'SystemGroup'}->removeMember($hGroups->{'Group'});
    DeleteAttributePermission($hSystemGroupAttributePermissions);
    ok( !ExistsAttributePermission($hSystemGroupAttributePermissions),
        'systemgroup attribute permissions were deleted' );
    DeleteAttributePermission($GroupAttributePermissions);
    ok( !ExistsAttributePermission($GroupAttributePermissions),
        'group attribute permissions were deleted' );

    return;
}

sub TestAllAttributePermissions {
    my ($Class, $Slogan, $hGroups) = @_;

    # retrieve all readable attributes for a given class and trustee
    foreach my $Trustee ( keys %{ $hGroups } ) {
        my $ahReadableAttributes = GetReadableAttributes($Class, $hGroups->{$Trustee});
        is( scalar @{ $ahReadableAttributes }, 0,
           'amount of readable attributes is 0 for class = ' .
            $Class->alias . ' and group = ' . $hGroups->{$Trustee}->alias
        );
    }

    # insert a read-permission on attribute 'Slogan' for everyone
    my $hSloganAttributePermission = {
        'AttributeID' => $Slogan->id,
        'TrusteeID'   => $hGroups->{'Everyone'}->id,
        'Permission'  => $ATTRIBUTE_PERMISSION_READ,
    };
    InsertAttributePermission($hSloganAttributePermission);

    foreach my $Trustee ( keys %{ $hGroups } ) {
        my $ahReadableAttributes = GetReadableAttributes($Class, $hGroups->{$Trustee});
        is( scalar @{ $ahReadableAttributes }, 1,
           'amount of readable attributes is 1 for class = ' .
            $Class->alias . ' and group = ' . $hGroups->{$Trustee}->alias
        );
        is( $ahReadableAttributes->[0]->{'AttributeID'}, $Slogan->id,
            'retrieved attribute id equals ' . $Slogan->id );
        is( $ahReadableAttributes->[0]->{'Permission'}, $ATTRIBUTE_PERMISSION_READ,
            'permissions are correctly set to ' . $ATTRIBUTE_PERMISSION_READ );
    }

    # create a subclass on FancyClass, add a new attribute
    my $OtherClass = $Class->insertSubClass({ 'Alias' => 'AnotherClass' });
    my $OtherAttribute = $OtherClass->insertAttribute({
        'Alias'   => 'HelloWorld',
        'Type'    => 'String',
        'Package' => 'DE_EPAGES::Object::API::Attributes::DefaultAttribute',
    });
    my $hOtherAttributePermission =  {
        'AttributeID' => $OtherAttribute->id,
        'TrusteeID'   => $hGroups->{'Everyone'}->id,
        'Permission'  => $ATTRIBUTE_PERMISSION_WRITE,
    };
    InsertAttributePermission($hOtherAttributePermission);

    foreach my $Trustee ( keys %{ $hGroups } ) {
        # subclass $OtherClass inherits also the 'Slogan' write permission from $Class
        my $ahReadableAttributes = GetReadableAttributes($OtherClass, $hGroups->{$Trustee});
        is( scalar @{ $ahReadableAttributes }, 2,
           'amount of readable attributes is 2 for class = ' .
            $OtherClass->alias . ' and group = ' . $hGroups->{$Trustee}->alias
        );
        my $ahWritableAttributes = GetWritableAttributes($OtherClass, $hGroups->{$Trustee});
        is( scalar @{ $ahWritableAttributes }, 1,
           'amount of writable attributes is 1 for class = ' .
            $OtherClass->alias . ' and group = ' . $hGroups->{$Trustee}->alias
        );
    }

    # cleanup
    DeleteAttributePermission($hSloganAttributePermission);
    ok( !ExistsAttributePermission($hSloganAttributePermission),
        'slogan attribute permissions were deleted' );
    DeleteAttributePermission($hOtherAttributePermission);
    ok( !ExistsAttributePermission($hOtherAttributePermission),
        'other attribute permissions were deleted' );

    return;
}

sub TestParameterPassing {
    my ($Class, $Attribute, $Group) = @_;

    my @Functions = ( \&IsAttributeReadable, \&IsAttributeWritable );
    foreach my $Function ( @Functions ) {
        my $Object = B::svref_2object $Function;
        my $FunctionName = $Object->GV->NAME;

        eval {
            $Function->(undef, undef);
        };
        ok( ExistsError() && GetError()->code eq 'PARAMETER_MISSING',
            "missing parameter (undef) in call to $FunctionName" );

        eval {
            $Function->('16380', 334);  ## no critic qw(ProhibitMagicNumbers)
        };
        ok( ExistsError() && GetError()->code eq 'PARAMETER_NO_OBJECT',
            "a string is no object in call to $FunctionName" );

        eval {
            $Function->($Attribute->id, undef);
        };
        ok( ExistsError() && GetError()->code eq 'PARAMETER_NO_OBJECT',
            "an attribute id is no object in call to $FunctionName" );

        eval {
            $Function->($Attribute, '334');
        };
        ok( ExistsError() && GetError()->code eq 'PARAMETER_NO_OBJECT',
            "a string as second parameter is no object in call to $FunctionName" );

        eval {
            $Function->($Attribute, $Group);
        };
        ok( !ExistsError(), "expected parameters passed to $FunctionName" );
    }


    @Functions = ( \&GetReadableAttributes, \&GetWritableAttributes );
    foreach my $Function ( @Functions ) {
        my $Object = B::svref_2object $Function;
        my $FunctionName = $Object->GV->NAME;

        eval {
            $Function->(undef, undef);
        };
        ok( ExistsError() && GetError()->code eq 'PARAMETER_MISSING',
            "missing parameter (undef) in call to $FunctionName" );

        eval {
            $Function->('16380', 334);  ## no critic qw(ProhibitMagicNumbers)
        };
        ok( ExistsError() && GetError()->code eq 'PARAMETER_NO_OBJECT',
            "a string is no object in call to $FunctionName" );

        eval {
            $Function->($Class->id, undef);
        };
        ok( ExistsError() && GetError()->code eq 'PARAMETER_NO_OBJECT',
            "a class id is no object in call to $FunctionName" );

        eval {
            $Function->($Class, '334');
        };
        ok( ExistsError() && GetError()->code eq 'PARAMETER_NO_OBJECT',
            "a string as second parameter is no object in call to $FunctionName" );

        eval {
            $Function->($Class, $Group);
        };
        ok( !ExistsError(), "expected parameters passed to $FunctionName" );
    }

    return;
}

sub Main {
    ok( Init(), 'initialization' );

    # create a user and some groups:
    my $System = LoadRootObject();
    my $TestParent =    $System->insertChild('User',  { 'Alias' => $TESTUSER    });
    my $RESTUser  = $TestParent->insertChild('User',  { 'Alias' => 'RESTUser'   });
    my $RESTGroup = $TestParent->insertChild('Group', { 'Alias' => 'RESTGroup'  });
    my $ParentGroup =   $System->insertChild('Group', { 'Alias' => $SYSTEMGROUP });

    # create a new class with some new attributes
    my $ObjectClass = LoadClassByAlias('Object');
    my $Class = $ObjectClass->insertSubClass({ 'Alias' => $CLASSNAME });

    my $Slogan = $Class->insertAttribute({
        'Alias'   => 'TestSlogan',
        'Type'    => 'String',
        'Package' => 'DE_EPAGES::Object::API::Attributes::DefaultAttribute',
    });

    my $ReadonlyAttribute = $Class->insertAttribute({
        'Alias'      => 'TestReadonly',
        'Type'       => 'String',
        'IsReadOnly' => '1',
        'Package'    => 'DE_EPAGES::Object::API::BaseAttribute',
    });

    #####################################################################################
    TestParameterPassing($Class, $Slogan, Everyone());
    TestAttributePermissions($Slogan, $RESTGroup);
    TestReadOnlyAttributePermissions($ReadonlyAttribute, $RESTGroup);
    TestAttributePermissionInheritance($Slogan, {
        'User'        => $RESTUser,
        'Group'       => $RESTGroup,
        'SystemGroup' => $ParentGroup,
    });
    TestAllAttributePermissions($Class, $Slogan, {
        'Everyone'    => Everyone(),
        'User'        => $RESTUser,
        'SystemGroup' => $ParentGroup,
    });
    #####################################################################################

    ok( Cleanup(), 'cleanup done' );

    return;
}

sub Init {
    Cleanup();

    return 1;
}

sub Cleanup {
    my $System = LoadRootObject();
    $System->child($TESTUSER)->delete if $System->existsChild($TESTUSER);
    $System->child($SYSTEMGROUP)->delete if $System->existsChild($SYSTEMGROUP);
    LoadClassByAlias($CLASSNAME)->delete if ExistsClassByAlias($CLASSNAME);

    return 1;
}

RunScript(
    Message => 'long',
    Sub     => sub {
        RunOnStore(
            Store => GetTestSite(),
            Sub   => \&Main,
        );
    },
);

1;