use strict;
use Test::More tests => 24;
use DE_EPAGES::Core::API::Script qw( RunScript );
use DE_EPAGES::Database::API::Connection qw( ConnectToSite DisconnectSite );
use DE_EPAGES::Object::API::Factory qw( GetRootObjectID LoadObject LoadObjectByPath
                                        ExistsClassByAlias LoadClassByAlias );
use DE_EPAGES::Permission::API::Permission qw( IsActionAllowed PERMISSION_ALLOW PERMISSION_DENY );
use DE_EPAGES::Permission::API::Table::Permission qw( InsertPermission );

sub Main {
    ConnectToSite();

    ok( Init(), 'initialization' );

    my $GroupEveryone = LoadObjectByPath( '/Groups/Everyone' );
    ok( $GroupEveryone->instanceOf( 'Group' ), 'group Everyone exists' );

    my $System = LoadObject( GetRootObjectID() );
    my $TestParent = $System->insertChild( 'User', { Alias => 'TestParent', Password=> '' } );
    my $User1  = $TestParent->insertChild( 'User',  { Alias => 'User1' , Password=> '' } );
    my $User2  = $TestParent->insertChild( 'User',  { Alias => 'User2' } );
    my $Group1 = $TestParent->insertChild( 'Group', { Alias => 'Group1' } );
    my $Group2 = $TestParent->insertChild( 'Group', { Alias => 'Group2' } );

    is( $User2->get('Password'), undef, 'empty password is stored as NULL' );
    is( $User1->get('Password'), undef, 'password can be NULL' );

    my $ClassTrustee = LoadClassByAlias( 'Trustee' );
    my $ClassGroup = LoadClassByAlias( 'Group' );
    my $ClassUser = LoadClassByAlias( 'User' );

    ok( $User1->instanceOf( $ClassUser ), 'user is User' );
    ok( $User1->instanceOf( $ClassTrustee), 'user is Trustee' );
    ok( !$User1->instanceOf( $ClassGroup), 'user is not Group' );

    ok( $Group1->instanceOf( $ClassGroup), 'group is Group' );
    ok( $Group1->instanceOf( $ClassTrustee), 'group is Trustee' );
    ok( !$Group1->instanceOf( $ClassUser), 'group is not User' );

    $Group1->addMember( $User1 );
    ok( $User1->isDirectMemberOf( $Group1 ), 'add user to group' );

    $Group1->addMember( $Group2 );
    ok( $Group2->isDirectMemberOf( $Group1 ), 'add group to group' );

    ok( $User1->isMemberOf( $Group1 ), 'direct group member' );

    $Group2->addMember( $User2 );
    ok( $User2->isMemberOf( $Group1 ), 'indirect group member' );

    ok( !$User1->isMemberOf( $Group2 ), 'member of outer group only' );

    $Group2->addMember( $Group1 );
    ok( $User1->isMemberOf( $Group2 ), 'mutual inclusion of groups' );
    ok( 2 == scalar @{$User1->groups}, 'groups count 1' );

    $Group1->removeMember( $User1 );
    ok( !$User1->isMemberOf( $Group2 ), 'remove group member' );
    ok( 0 == scalar @{$User1->groups}, 'groups count 2' );

    # permissions
    my $ClassObject = LoadClassByAlias( 'Object' );
    my $ObjectClass1 = $ClassObject->insertSubClass( { Alias => 'C1' } );
    my $Object1 = $TestParent->insertChild( 'C1', { Alias => 'Object1' } );
    my $ActionView = $ObjectClass1->folder('Actions')->insertChild('Action', { Alias => 'View', ActionClass=> $ObjectClass1, 'Package'=>'main'  } );
    my $RoleManager = $ObjectClass1->folder('Roles')->insertChild('Role', { Alias => 'Manager', ActionClass=> $ObjectClass1  } );
    $RoleManager->addAction( $ActionView );

    ok( !IsActionAllowed( $Object1, $User2, $ActionView ), 'before InsertPermission' );

    InsertPermission( {
        ObjectID => $Object1->id,
        TrusteeID => $Group1->id,
        MetaActionID => $RoleManager->id,
        Allow => PERMISSION_ALLOW
    } );

    ok( IsActionAllowed( $Object1, $User2, $ActionView ), 'permission granted' );
    ok( !IsActionAllowed( $Object1, $User1, $ActionView ), 'no permissions set' );

    InsertPermission( {
        ObjectID => $Object1->id,
        TrusteeID => $Group2->id,
        MetaActionID => $RoleManager->id,
        Allow => PERMISSION_DENY
    } );

    ok( !IsActionAllowed( $Object1, $User2, $ActionView ), 'permission actively denied' );

    ok( Cleanup(), 'cleanup' );

    DisconnectSite();
}

sub Cleanup {

    my $System = LoadObject( GetRootObjectID() );
    $System->child('TestParent')->delete if $System->existsChild( 'TestParent' );

    foreach my $ClassName ( qw( C1 ) ) {
        next unless ExistsClassByAlias( $ClassName );
        LoadClassByAlias( $ClassName )->delete;
    }

    return 1;
}

sub Init {
    Cleanup();

    return 1;
}

RunScript(Sub=>\&Main, Debug=>1);
